import logging
from ogip.tools import convolve
import os
import numpy as np
import pytest

import ogip.spec
from matplotlib import pyplot as plt   

logging.basicConfig(level=logging.DEBUG)


def test_crab_fit(crab_spectra):
    crab_pha_fn, crab_rmf_fn, crab_arf_fn = crab_spectra
    
    crab_pha = ogip.spec.PHAI.from_file_name(crab_pha_fn)
    crab_rmf = ogip.spec.RMF.from_file_name(crab_rmf_fn)
    crab_arf = ogip.spec.ARF.from_file_name(crab_arf_fn)
            
    from ogip.tools import plot, transform_rmf, crab_ph_cm2_s_kev    
    
    def gen_line(E0, depth):
        return lambda x: (x/20)**-2 * (1 - depth * np.exp(-(x-E0)**2/(E0*0.1)**2)/2)

    for E0, rmf_bias, depth in [
        (25, 1, 1),
        (60, 5, 1),
    ]:
        f = plt.figure(figsize=(15, 10))
        
        t_rmf = transform_rmf(crab_rmf, 
                            crab_arf, 
                            lambda x:(x+rmf_bias), 
                            preserved_projection=crab_ph_cm2_s_kev)

        plot(crab_pha, crab_ph_cm2_s_kev, crab_rmf, crab_arf, fig=f, label_prefix="crab")    
        plot(None, crab_ph_cm2_s_kev, t_rmf, crab_arf, fig=f, label_prefix="transformed, crab")

        # Crab spectrum is the same with new and old RMF
        s = convolve(crab_ph_cm2_s_kev, crab_rmf, crab_arf) 
        t_s = convolve(crab_ph_cm2_s_kev, t_rmf, crab_arf)
        assert np.all(np.abs((s - t_s)/s) < 0.001)

        plot(None, gen_line(E0, depth), crab_rmf, crab_arf, fig=f, label_prefix="line")
        plot(None, gen_line(E0, depth), t_rmf, crab_arf, fig=f, label_prefix="transformed, line", 
             plot_kwargs=dict(lw=5, alpha=0.8))
        plot(None, gen_line(E0 - rmf_bias, depth), crab_rmf, crab_arf, fig=f, label_prefix="shifted line")

        plt.savefig(f"conv_{E0}.png")
        plt.savefig("conv.png")
        
        # RMF transformation is equivalent to shifting. The match is not exact, but quite close
        s = convolve(gen_line(E0, depth), crab_rmf, crab_arf) 
        t_s = convolve(gen_line(E0 - rmf_bias, depth), t_rmf, crab_arf)
        
        if E0 < 50: # else shape too affected
            assert np.nanstd(np.abs((s - t_s)/s)) < 0.05

        t_rmf.to_fits("transformed_rmf.fits")

        


import os
import pytest

@pytest.fixture
def crab_spectra():
    crab_pha_fn = 'crab_pha.fits'
    crab_rmf_fn = 'crab_rmf.fits'
    crab_arf_fn = 'crab_arf.fits'
    
    if os.path.exists(crab_pha_fn) and os.path.exists(crab_rmf_fn) and os.path.exists(crab_arf_fn):
        pass
    else:
        import oda_api.api
        disp = oda_api.api.DispatcherAPI()
        prod = disp.get_product(
            instrument='isgri',
            product='isgri_spectrum',
            scw_list="066500220010.001",
            osa_version="OSA10.2"
            )

        
        prod.isgri_spectrum_0_Crab_isgri_spectrum.write_fits_file(crab_pha_fn)
        prod.isgri_spectrum_1_Crab_isgri_arf.write_fits_file(crab_arf_fn)
        prod.isgri_spectrum_2_Crab_isgri_rmf.write_fits_file(crab_rmf_fn)

    return crab_pha_fn, crab_rmf_fn, crab_arf_fn


import logging
import os
import numpy as np
import pytest

import ogip.spec

logging.basicConfig(level=logging.DEBUG)


def test_read_something():
    import ogip.core

    for fn, c in [
                ("tests/data/phaI.fits.gz", ogip.spec.PHAI),
                ("tests/data/rmf_rt16_116.fits", ogip.spec.RMF),
                ("tests/data/rmf.fits.gz", ogip.spec.RMF),
                ("tests/data/arf.fits.gz", ogip.spec.ARF),
            ]:
        o = ogip.core.open_something(fn)
        assert isinstance(o, c)
        print(o.to_long_string())


def synth_e():
    return (np.logspace(1, 3, 100),
            np.logspace(1, 3, 50))

def synth_phaI():
    elh, eb = synth_e()

    r = np.random.normal(size=eb[:-1].shape)*100

    phaI = ogip.spec.PHAI.from_arrays(
                1.,
                rate=r,
                stat_err=r/100.
            )

    return phaI

def synth_rmf():
    elh, eb = synth_e()

    m_e, m_eb = np.meshgrid(elh[:-1],eb[:-1])

    d = (m_e-m_eb)/(0.1*m_e)
    matrix = np.zeros_like(d)

    m = d<10
    matrix[m] = np.exp(-d**2)[m]

    rmf = ogip.spec.RMF.from_arrays(
                energ_lo = elh[:-1],
                energ_hi = elh[1:],
                matrix = np.transpose(matrix),
                e_min = eb[:-1],
                e_max = eb[1:],
            )

    return rmf


def test_bad_arrays():
    pass

def test_from_arrays():

    rmf = synth_rmf()

    rmf.to_fits("rmf.fits")

@pytest.mark.skipif(not os.environ.get('HEADAS', None), reason="need HEASoft to test reading with xspec")
def test_read_xspec():
    import xspec

    synth_phaI().to_fits("phaI.fits")
    synth_rmf().to_fits("rmf.fits")

    s = xspec.Spectrum("phaI.fits")

    s.response = "rmf.fits"



import numpy as np
from ogip.spec import PHAI, RMF, ARF

def convolve(source_model: callable, rmf: RMF, arf: ARF=None):
    ie1 = rmf._energ_lo
    ie2 = rmf._energ_hi
    e1 = rmf._e_min
    e2 = rmf._e_max
    
    if arf is None:
        _arf = np.ones_like(ie1)
    else:
        _arf = arf._arf    

    return (np.outer(_arf * source_model(ie1)*(ie2-ie1),np.ones_like(e1))*rmf._matrix).sum(0)

def plot(pha: PHAI, model: callable, rmf: RMF, arf: ARF=None, fig=None, label_prefix=None, plot_kwargs={}):
    import matplotlib as mpl
    mpl.use('Agg')
    import matplotlib.pylab as plt

    ie1 = rmf._energ_lo
    ie2 = rmf._energ_hi
    e1 = rmf._e_min
    e2 = rmf._e_max

    model_spec = convolve(model, rmf, arf)

    def prefix(l):
        if label_prefix is None:
            return l
        else:
            return f"{label_prefix}: {l}"

    if fig is None:
        plt.figure(figsize=(15,10))

    plt.step(
        e1,
        model_spec/(e2 - e1),
        where='post',
        label=prefix("model"),
        **plot_kwargs
    )
    
    if pha is not None:
        x = plt.step(
            e1,
            pha._rate/(e2 - e1),
            where='post',
            label=prefix("data"),
            **plot_kwargs
        )
                
        plt.errorbar(
            (e1 * e2)**0.5,
            pha._rate/(e2 - e1),
            pha._stat_err/(e2 - e1),
            ls="",
            c=x[0].get_color(),
            **plot_kwargs
        )


    plt.legend()     
    plt.loglog()
    
        
    plt.xlim([15, 600])
    plt.grid()

def transform_rmf(rmf: RMF, arf: ARF, bias_function: callable, preserved_projection=None, tolerance=0.05) -> RMF:
    new_rmf = RMF.from_arrays(
        energ_lo=bias_function(rmf._energ_lo.copy()),
        energ_hi=bias_function(rmf._energ_hi.copy()),
        e_min=rmf._e_min.copy(),
        e_max=rmf._e_max.copy(),
        matrix=rmf._matrix.copy()
    )

    if preserved_projection is not None:
        # projection on model should be preserved
        corr = convolve(preserved_projection, new_rmf, arf) / convolve(preserved_projection, rmf, arf)

        new_rmf._matrix /= np.outer(np.ones_like(new_rmf._energ_lo), corr)

        s = convolve(preserved_projection, rmf, arf) 
        t_s = convolve(preserved_projection, new_rmf, arf)

        assert np.all(np.abs((s - t_s)/s) < tolerance)
    

    return new_rmf

# Roques & Jourdain 2018
def crab_ph_cm2_s_kev(en):
    K=7.417e-4
    al1=-1.98
    al2=-2.33
    Ec=500.
    f=K*(en/100)**al1*(np.exp(-en/Ec))
    m=en>Ec*(al1-al2)
    f[m]=(K*((al1-al2)*Ec/100)**(al1-al2)*(en/100)**al2*np.exp(-(al1-al2)))[m]

    return f

def le_bias_function(energy, bias_keV):
    # TODO: preserves 60 and 511, shifts LE
    return energy + bias_keV